
#include <math.h>
#include <sysexits.h>
#include <IOKit/IOCFPlugIn.h>
#include <IOKit/hid/IOHIDLib.h>

#include "input.h"
#include "ar.h"

extern const double NSAppKitVersionNumber;

int ar_codes_tiger[] = {
	[25] = { AR_PREV_HOLD }, // 14+6+3+2
	[26] = { AR_NEXT_HOLD }, // 14+6+4+2
	[40] = { AR_MENU_HOLD }, // 14+6+14+6
	[43] = { AR_VUP },       // 14+12+11+6
	[44] = { AR_VDOWN },     // 14+13+11+6
	[50] = { AR_PREV },      // 14+10+6+14+10+6
	[54] = { AR_MENU },      // 14+7+6+14+7+6
	[56] = { AR_PLAY },      // 14+8+6+14+8+6
	[58] = { AR_NEXT },      // 14+9+6+14+9+6
	[76] = { AR_PLAY_HOLD }, // 18+14+6+18+14+6
};

int ar_codes_leopard[] = {
	[74]  = { AR_NEXT_HOLD }, // 31+19+18+4+2
	[73]  = { AR_PREV_HOLD }, // 31+19+18+3+2
	[125] = { AR_VUP },       // 31+29+28+19+18
	[126] = { AR_VDOWN },     // 31+30+28+19+18
	[136] = { AR_MENU_HOLD }, // 31+19+18+31+19+18
	[176] = { AR_MENU },      // 31+20+19+18+31+20+19+18
	[178] = { AR_PLAY },      // 31+21+19+18+31+21+19+18
	[180] = { AR_NEXT },      // 31+22+19+18+31+22+19+18
	[182] = { AR_PREV },      // 31+23+19+18+31+23+19+18
	[206] = { AR_PLAY_HOLD }, // 35+31+19+18+35+31+19+18
};

int *ar_codes = &ar_codes_tiger;

static IOHIDQueueInterface **queue;
static IOHIDDeviceInterface	**hidDeviceInterface = NULL;
static int inited = 0;
static io_iterator_t hidObjectIterator = (io_iterator_t)NULL;

static int FindHIDDevices(mach_port_t masterPort, io_iterator_t *hidObjectIterator)
{
    CFMutableDictionaryRef hidMatchDictionary = NULL;
    IOReturn ioReturnValue = kIOReturnSuccess;
    
    // Set up a matching dictionary to search the I/O Registry by class
	// name for all HID class devices
	hidMatchDictionary = IOServiceMatching("AppleIRController");

 	// Now search I/O Registry for matching devices.
 	ioReturnValue = IOServiceGetMatchingServices(masterPort, hidMatchDictionary, hidObjectIterator);

	//If search is unsuccessful, print message and hang.
	if ((ioReturnValue != kIOReturnSuccess) | (*hidObjectIterator == NULL)) {
		return -1;
	}
	return 0;
}

IOHIDElementCookie *cookies = NULL;
int nr_cookies = 0;

static IOHIDElementCookie *getHIDCookies(IOHIDDeviceInterface122 **handle)
{
	IOHIDElementCookie *cookies = calloc(256, sizeof(IOHIDElementCookie));
	CFTypeRef object;
	long number;
	CFArrayRef elements;
	CFDictionaryRef element;
	CFIndex i;

	if (!handle || !(*handle)) {
		free(cookies);
		return NULL;
	}

	// Copy all elements, since we're grabbing most of the elements
	// for this device anyway, and thus, it's faster to iterate them
	// ourselves. When grabbing only one or two elements, a matching
	// dictionary should be passed in here instead of NULL.
	if (((*handle)->copyMatchingElements(handle, NULL, &elements)) != kIOReturnSuccess) {
		free(cookies);
		return NULL;
	}

	for (i=0; i<CFArrayGetCount(elements); i++) {
		element = CFArrayGetValueAtIndex(elements, i);

		//Get cookie
		object = (CFDictionaryGetValue(element, CFSTR(kIOHIDElementCookieKey)));
		if (object == 0 || CFGetTypeID(object) != CFNumberGetTypeID())
			continue;
		if(!CFNumberGetValue((CFNumberRef) object, kCFNumberLongType, &number))
			continue;
		cookies[nr_cookies++] = (int)number;
	}

	return cookies;
}

static int CreateHIDDeviceInterface(io_object_t hidDevice, IOHIDDeviceInterface ***hidDeviceInterface)
{
	io_name_t className;
	IOCFPlugInInterface **plugInInterface = NULL;
	SInt32 score = 0;

	if (IOObjectGetClass(hidDevice, className) != kIOReturnSuccess)
		return -1;

	if (IOCreatePlugInInterfaceForService(hidDevice, kIOHIDDeviceUserClientTypeID, kIOCFPlugInInterfaceID, &plugInInterface, &score) != kIOReturnSuccess)
		return -1;

	//Call a method of the intermediate plug-in to create the device interface
	if ((*plugInInterface)->QueryInterface(plugInInterface, CFUUIDGetUUIDBytes(kIOHIDDeviceInterfaceID), (LPVOID) hidDeviceInterface) != S_OK)
		return -1;
	(*plugInInterface)->Release(plugInInterface);

	return 0;
}

int mp_input_ar_init(void)
{
	io_object_t	hidDevice = (io_object_t)NULL;
	int i;
	
	if (nr_cookies) {
		if (cookies)
			free(cookies);
		nr_cookies = 0;
	}
	
	if (floor(NSAppKitVersionNumber) <= 824 /* NSAppKitVersionNumber10_4 */) {
		ar_codes = &ar_codes_tiger;
	} else {
		ar_codes = &ar_codes_leopard;
	}

	if (FindHIDDevices(kIOMasterPortDefault, &hidObjectIterator))
		return -1;
	if (hidObjectIterator == (io_iterator_t)NULL)
		return -1;

	while ((hidDevice = IOIteratorNext(hidObjectIterator))) {
		if (CreateHIDDeviceInterface(hidDevice, &hidDeviceInterface) < 0)
			return -1;
		if ((cookies = getHIDCookies((IOHIDDeviceInterface122 **)hidDeviceInterface)) == NULL)
			return -1;
		IOObjectRelease(hidDevice);
		if (hidDeviceInterface != NULL)
			break;
	}
	if (hidDeviceInterface == NULL)
			return -1;

	//open the device
	if ((*hidDeviceInterface)->open(hidDeviceInterface, kIOHIDOptionsTypeSeizeDevice) != kIOReturnSuccess)
		return -1;

	if ( (queue = (*hidDeviceInterface)->allocQueue(hidDeviceInterface)) == NULL)
		return -1;

	//create the queue
	/* depth: maximum number of elements in queue before oldest elements in queue begin to be lost. */
	(*queue)->create(queue, 0, 12);

	//add elements to the queue
	for (i = 0;i < nr_cookies;i++)
		(*queue)->addElement(queue, cookies[i], 0);

	//start data delivery to queue
	(*queue)->start(queue);
	inited = 1;
	return 0;
}

int mp_input_ar_read(int fd)
{
	int down = 0;
	int ret = MP_INPUT_NOTHING;
	AbsoluteTime zeroTime = {0,0};
	IOHIDEventStruct event;
	static int prev_event = 0;
	IOReturn result = kIOReturnSuccess;

	int cookie = 0;
	int value = 0;

	if (inited == 0)
		return MP_INPUT_NOTHING;

	while (result == kIOReturnSuccess) {
		if ((result = (*queue)->getNextEvent(queue, &event, zeroTime, 0)) != kIOReturnSuccess)
			continue;
		if (event.elementCookie == 5)
			continue;
#ifdef TEST
		printf(" - event cookie: %d, value: %d, long value: %d\n", event.elementCookie, event.value, event.longValue);
#endif
		cookie += event.elementCookie;
		value += event.value;
	}
	
	if (cookie) {
#ifdef TEST
		printf(" ! event: %d, value: %d\n", cookie, value);
#endif
		ret = ar_codes[cookie];
		switch (ret) {
			case AR_VUP:
			case AR_VDOWN:
			case AR_NEXT_HOLD:
			case AR_PREV_HOLD:
				down = (value ? 1 : 0) * MP_KEY_DOWN;
				break;
			
			default:
				down = 0;
		}
	}

mp_input_ar_read_exit:
	if (ret > 0)
		prev_event = ret;
	return ret | down;
}

void mp_input_ar_close(int fd)
{
	if (inited == 0)
		return;

	//close the device
	(*hidDeviceInterface)->close(hidDeviceInterface);
	//release the interface
	(*hidDeviceInterface)->Release(hidDeviceInterface);
	//stop data delivery to queue
	(*queue)->stop(queue);
	//dispose of queue
	(*queue)->dispose(queue);
	//release the queue we allocated
//	(*queue)->Release(queue);
	(*hidDeviceInterface)->Release(hidDeviceInterface);
	//Release iterator. Don't need to release iterator objects.
	IOObjectRelease(hidObjectIterator);
}

#ifdef TEST
int main(void)
{
	int ret;

	if (mp_input_ar_init() < 0) {
		printf("Unable to initialize Apple Remote.\n");
		return 1;
	}

	while (1) {
		switch ((ret = mp_input_ar_read(0)) & ~MP_KEY_DOWN) {
			case AR_PLAY:       printf(" - AR_PLAY."); break;
			case AR_PLAY_HOLD:  printf(" - AR_PLAY_HOLD."); break;
			case AR_NEXT:       printf(" - AR_NEXT."); break;
			case AR_NEXT_HOLD:  printf(" - AR_NEXT_HOLD."); break;
			case AR_PREV:       printf(" - AR_PREV."); break;
			case AR_PREV_HOLD:  printf(" - AR_PREV_HOLD."); break;
			case AR_MENU:       printf(" - AR_MENU."); break;
			case AR_MENU_HOLD:  printf(" - AR_MENU_HOLD."); break;
			case AR_VUP:        printf(" - AR_VUP."); break;
			case AR_VDOWN:      printf(" - AR_VDOWN."); break;
		}
		if ((ret > 0 )&&(ret & MP_KEY_DOWN))
			printf(" [hold]");
		if (ret > 0)
			printf("\n");
	}

	mp_input_ar_close(0);
	return 0;
}
#endif
